/*
 * I generate a table of positions on an sscan record to allow a scan of three
 * regions with a different step size for each region.
 *
 * Macro parameters:
 *   P        PV name prefix (e.g. "17bm:")
 */

program regScanPosGen

option +r;

%%#include <string.h>
%%#include <stdlib.h>
%%#include <math.h>
%%#include <errlog.h>
%%#include <epicsAssert.h>
%%#include <epicsStdio.h>
%%#include "regScanPosGenLib.h"

%{
  static int fequal(double value1, double value2, double errorLimit);
  static void logMessage(SS_ID ssId, struct UserVar *pVar,
      const char *message);
  static void zeroDoubleArray(double array[], size_t arraySize);
}%

#define SNL_NULL 0
#define SNCSEQ_PV_STAT_OK 0 /* Must match pvStatOK in include/pvAlarm.h */

#define VARIABLE_ASSIGNMENT_TIMEOUT 5.0 /* in seconds */
#define POSITIONS_SIZE 2000
#define ERROR_MESSAGE_SIZE 1024
#define POSITIONS_PV_NAME_SIZE 1024
#define NUMBER_OF_POINTS_PV_NAME_SIZE 1024

#define GO_DONE 0
#define GO_GENERATE 1

#define STATUS_OK 0
#define STATUS_GENERATING 1
#define STATUS_ERROR 2

#define PV_PREFIX_KEY "P"

char *PvPrefix;
int HadError;
char ErrorMessage[ERROR_MESSAGE_SIZE];
char PositionsPvName[POSITIONS_PV_NAME_SIZE];
char NumberOfPointsPvName[NUMBER_OF_POINTS_PV_NAME_SIZE];
char *ScanRecordName[5]; /* must be consistent with ScanPosGenTargetScan in
                            regScanPosGen.db */
char *ScanPositionerName[4]; /* must be consistent with ScanPosGenTargetPos in
                                regScanPosGen.db */
int PvStatus;
int PvStatus2;
int Index;
int NumCharsPrinted;

/* Control variables */

short Go;
  assign Go to "{P}ScanPosGenGo";
  monitor Go;
  evflag GoChanged;
  sync Go to GoChanged;

short Status;
  assign Status to "{P}ScanPosGenStatus";

/* Parameter variables */

double StartOfRegion[3];
  assign StartOfRegion to {
    "{P}ScanPosGenReg1Start",
    "{P}ScanPosGenReg2Start",
    "{P}ScanPosGenReg3Start"
  };

double EndOfRegion[3];
  assign EndOfRegion to {
    "{P}ScanPosGenReg1End",
    "{P}ScanPosGenReg2End",
    "{P}ScanPosGenReg3End"
  };

double StepSizeOfRegion[3];
  assign StepSizeOfRegion to {
    "{P}ScanPosGenReg1StepSz",
    "{P}ScanPosGenReg2StepSz",
    "{P}ScanPosGenReg3StepSz"
  };

short TargetScan;
  assign TargetScan to "{P}ScanPosGenTargetScan";
short TargetPositioner;
  assign TargetPositioner to "{P}ScanPosGenTargetPos";

double Positions[POSITIONS_SIZE];
  assign Positions to ""; /* dynamically assigned at run-time to the target
                           * positioner array (e.g. 17bm:scan1.P1PA) */
int PositionsSize;
int NumberOfPoints;
  assign NumberOfPoints to ""; /* dynamically assigned at run-time to the
                                * target NPTS field (e.g. 17bm:scan1.NPTS) */

/* Support variables */


ss Main {
  state initialize {
    when () {
      PvPrefix = macValueGet(PV_PREFIX_KEY);
      assert(PvPrefix != SNL_NULL);

      ScanRecordName[0] = "scan1";
      ScanRecordName[1] = "scan2";
      ScanRecordName[2] = "scan3";
      ScanRecordName[3] = "scan4";
      ScanRecordName[4] = "scanH";

      ScanPositionerName[0] = "P1PA";
      ScanPositionerName[1] = "P2PA";
      ScanPositionerName[2] = "P3PA";
      ScanPositionerName[3] = "P4PA";
    } state watchForGenerateRequest
  }

  state watchForGenerateRequest {
    when (efTestAndClear(GoChanged) && Go == GO_GENERATE) {
      Status = STATUS_GENERATING;
      pvPut(Status);
    } state handleGenerateRequest
  }

  state handleGenerateRequest {
    when () {} state getGenerationParameters
  }

  state getGenerationParameters {
    when () {
      PvStatus = SNCSEQ_PV_STAT_OK;
      HadError = 0;

      for (Index = 0; Index < 3; Index++) {
        if (PvStatus == SNCSEQ_PV_STAT_OK) {
          PvStatus = pvGet(StartOfRegion[Index]);
        }
        if (PvStatus == SNCSEQ_PV_STAT_OK) {
          PvStatus = pvGet(EndOfRegion[Index]);
        }
        if (PvStatus == SNCSEQ_PV_STAT_OK) {
          PvStatus = pvGet(StepSizeOfRegion[Index]);
        }
      }

      if (PvStatus == SNCSEQ_PV_STAT_OK) PvStatus = pvGet(TargetScan);
      if (PvStatus == SNCSEQ_PV_STAT_OK) PvStatus = pvGet(TargetPositioner);

      if (PvStatus != SNCSEQ_PV_STAT_OK) {
        HadError = 1;
        strncpy(ErrorMessage, "failed to get at least one of generation "
          "parameters", ERROR_MESSAGE_SIZE);
        %%logMessage(ssId, pVar, pVar->ErrorMessage);
      }
    } state checkGetGenerationParametersForError
  }
  state checkGetGenerationParametersForError {
    when (!HadError) {} state assignPositionsAndNumberOfPoints
    when (HadError) {
      Status = STATUS_ERROR;
      pvPut(Status);
    } state done
  }

  state assignPositionsAndNumberOfPoints {
    when () {
      %{
        pVar->NumCharsPrinted = epicsSnprintf(pVar->PositionsPvName,
            POSITIONS_PV_NAME_SIZE, "%s%s.%s", pVar->PvPrefix,
            pVar->ScanRecordName[pVar->TargetScan],
            pVar->ScanPositionerName[pVar->TargetPositioner]);
      }%
      if (NumCharsPrinted < POSITIONS_PV_NAME_SIZE) {
        pvAssign(Positions, PositionsPvName);
      } else {
        HadError = 1;
        strncpy(ErrorMessage, "failed to create positions array PV name",
          ERROR_MESSAGE_SIZE);
        %%logMessage(ssId, pVar, pVar->ErrorMessage);
      }

      %{
        pVar->NumCharsPrinted = epicsSnprintf(pVar->NumberOfPointsPvName,
            NUMBER_OF_POINTS_PV_NAME_SIZE, "%s%s.NPTS", pVar->PvPrefix,
            pVar->ScanRecordName[pVar->TargetScan]);
      }%
      if (NumCharsPrinted < NUMBER_OF_POINTS_PV_NAME_SIZE) {
        pvAssign(NumberOfPoints, NumberOfPointsPvName);
      } else {
        HadError = 1;
        strncpy(ErrorMessage, "failed to create number of points PV name",
          ERROR_MESSAGE_SIZE);
        %%logMessage(ssId, pVar, pVar->ErrorMessage);
      }
    } state checkAssignPositionsAndNumberOfPointsForError
  }
  state checkAssignPositionsAndNumberOfPointsForError {
    when (!HadError) {} state waitForAssignmentsToComplete
    when (HadError) {
      Status = STATUS_ERROR;
      pvPut(Status);
    } state done
  }

  state waitForAssignmentsToComplete {
    when (pvConnectCount() == pvAssignCount()) {
    } state generatePositions
    when (delay(VARIABLE_ASSIGNMENT_TIMEOUT)) {
      strncpy(ErrorMessage, "timed out waiting for assignment of positions "
        "and number of points variables to target PVs", ERROR_MESSAGE_SIZE);
      %%logMessage(ssId, pVar, pVar->ErrorMessage);
      Status = STATUS_ERROR;
      pvPut(Status);
    } state done
  }

  state generatePositions {
    when () {
      %%int positionsIndex;
      %%int numPointsGenerated;
      %%int i;

      %%positionsIndex = 0;
      %%numPointsGenerated = 0;

      zeroDoubleArray(Positions, POSITIONS_SIZE);

      PositionsSize = pvCount(Positions);

      %{
        for (i = 0; i < 3; i++) {
          if (i > 0 && positionsIndex > 0 &&
              fequal(pVar->Positions[positionsIndex-1],
                pVar->StartOfRegion[i], 1e-5)) {
            positionsIndex--;
          }
          numPointsGenerated = imcaGenerateScanPositions(pVar->Positions,
            pVar->PositionsSize, positionsIndex, pVar->StartOfRegion[i],
            pVar->EndOfRegion[i], pVar->StepSizeOfRegion[i]);
          positionsIndex += numPointsGenerated;
        }
        pVar->NumberOfPoints = positionsIndex;
      }%

      PvStatus = pvPut(Positions, SYNC);
      PvStatus2 = pvPut(NumberOfPoints, SYNC);

      if (PvStatus == SNCSEQ_PV_STAT_OK && PvStatus2 == SNCSEQ_PV_STAT_OK) {
        Status = STATUS_OK;
      } else {
        strncpy(ErrorMessage, "failed to set positions and number of points",
          ERROR_MESSAGE_SIZE);
        %%logMessage(ssId, pVar, pVar->ErrorMessage);
        Status = STATUS_ERROR;
      }
      pvPut(Status);
    } state done
  }

  state done {
    when () {
      Go = GO_DONE;
      pvPut(Go);
    } state watchForGenerateRequest
  }
}

%{
static int fequal(double value1, double value2, double errorLimit) {
  return fabs(value1 - value2) < errorLimit;
}

static void logMessage(SS_ID ssId, struct UserVar *pVar,
    const char *message) {
  errlogPrintf("%sregScanPosGen: %s\n", pVar->PvPrefix, message);
}

static void zeroDoubleArray(double array[], size_t arraySize) {
  int i;

  for (i = 0; i < arraySize; i++) array[i] = 0.0;
}
}%
