#include <iocsh.h>
#include <registryFunction.h>
#include "imca-epics-unit.h"
#include "regScanPosGenLib.h"
#include <epicsExport.h>

static EpicsUnitContext euc;

static void testImcaGenerateScanPositions01(void) {
  /* zero size positions array */
  double positions[0];
  int numGen;

  numGen = imcaGenerateScanPositions(positions, 0, 0, 0, 5, 1);
  failIfIntNotEqual(&euc, 0, numGen);
}
static void testImcaGenerateScanPositions02(void) {
  /* size=1 positions array */
  double positions[1];
  int numGen;

  numGen = imcaGenerateScanPositions(positions, 1, 0, 2, 5, 1);
  failIfIntNotEqual(&euc, 1, numGen);
  failIfDoubleNotEqual(&euc, 2.0, positions[0], 0.0);
}
static void testImcaGenerateScanPositions03(void) {
  /* size=2 positions array */
  double positions[2];
  int numGen;

  numGen = imcaGenerateScanPositions(positions, 2, 0, 2, 5, 1);
  failIfIntNotEqual(&euc, 2, numGen);
  failIfDoubleNotEqual(&euc, 2.0, positions[0], 0.0);
  failIfDoubleNotEqual(&euc, 3.0, positions[1], 0.0);
}
static void testImcaGenerateScanPositions04(void) {
  /* basic */
  int positionsSize;
  double start;
  double end;
  double stepSize;
  double positions[6];
  int numGen;
  double position;
  int i;

  positionsSize = sizeof(positions) / sizeof(positions[0]);
  start = 1.0;
  end = 2.0;
  stepSize = 0.2;

  numGen = imcaGenerateScanPositions(positions, positionsSize, 0, start, end,
    stepSize);
  failIfIntNotEqual(&euc, positionsSize, numGen);
  position = start;
  for (i = 0; i < positionsSize; i++) {
    failIfDoubleNotEqual(&euc, position, positions[i], 0.0);
    position += stepSize;
  }
}
static void testImcaGenerateScanPositions05(void) {
  /* basic */
  int positionsSize;
  double start;
  double end;
  double stepSize;
  double positions[500];
  int numGen;
  double position;
  int i;

  positionsSize = sizeof(positions) / sizeof(positions[0]);
  start = 0.0;
  end = 100.0;
  stepSize = 0.2;

  numGen = imcaGenerateScanPositions(positions, positionsSize, 0, start, end,
    stepSize);
  failIfIntNotEqual(&euc, positionsSize, numGen);
  position = start;
  for (i = 0; i < positionsSize; i++) {
    failIfDoubleNotEqual(&euc, position, positions[i], 0.0);
    position += stepSize;
  }
}
static void testImcaGenerateScanPositions06(void) {
  /* last step would go over end */
  int positionsSize;
  double start;
  double end;
  double stepSize;
  double positions[5];
  int numGen;

  positionsSize = sizeof(positions) / sizeof(positions[0]);
  start = 0.0;
  end = 0.5;
  stepSize = 0.2;

  numGen = imcaGenerateScanPositions(positions, positionsSize, 0, start, end,
    stepSize);
  failIfIntNotEqual(&euc, 3, numGen);
  failIfDoubleNotEqual(&euc, 0.0, positions[0], 0.0);
  failIfDoubleNotEqual(&euc, 0.2, positions[1], 0.0);
  failIfDoubleNotEqual(&euc, 0.4, positions[2], 0.0);
}
static void testImcaGenerateScanPositions07(void) {
  /* negative start and end */
  int positionsSize;
  double start;
  double end;
  double stepSize;
  double positions[5];
  int numGen;

  positionsSize = sizeof(positions) / sizeof(positions[0]);
  start = -4.0;
  end = -2.0;
  stepSize = 1.0;

  numGen = imcaGenerateScanPositions(positions, positionsSize, 0, start, end,
    stepSize);
  failIfIntNotEqual(&euc, 3, numGen);
  failIfDoubleNotEqual(&euc, -4.0, positions[0], 0.0);
  failIfDoubleNotEqual(&euc, -3.0, positions[1], 0.0);
  failIfDoubleNotEqual(&euc, -2.0, positions[2], 0.0);
}
static void testImcaGenerateScanPositions08(void) {
  /* negative step */
  int positionsSize;
  double start;
  double end;
  double stepSize;
  double positions[5];
  int numGen;

  positionsSize = sizeof(positions) / sizeof(positions[0]);
  start = 4.0;
  end = 2.0;
  stepSize = -1.0;

  numGen = imcaGenerateScanPositions(positions, positionsSize, 0, start, end,
    stepSize);
  failIfIntNotEqual(&euc, 3, numGen);
  failIfDoubleNotEqual(&euc, 4.0, positions[0], 0.0);
  failIfDoubleNotEqual(&euc, 3.0, positions[1], 0.0);
  failIfDoubleNotEqual(&euc, 2.0, positions[2], 0.0);
}
static void testImcaGenerateScanPositions09(void) {
  /* end less than start */
  int positionsSize;
  double start;
  double end;
  double stepSize;
  double positions[5];
  int numGen;

  positionsSize = sizeof(positions) / sizeof(positions[0]);
  start = 4.0;
  end = 2.0;
  stepSize = 1.0;

  numGen = imcaGenerateScanPositions(positions, positionsSize, 0, start, end,
    stepSize);
  failIfIntNotEqual(&euc, 1, numGen);
  failIfDoubleNotEqual(&euc, 4.0, positions[0], 0.0);
}
static void testImcaGenerateScanPositions10(void) {
  /* basic */
  int expectedSize;
  int positionsSize;
  double expected[] = { 0.0, 3.0, 6.0, 9.0 };
  double positions[10];
  int numGen;

  expectedSize = sizeof(expected) / sizeof(expected[0]);
  positionsSize = sizeof(positions) / sizeof(positions[0]);

  numGen = imcaGenerateScanPositions(positions, positionsSize, 0, 0.0, 10.0,
    3.0);
  failIfIntNotEqual(&euc, expectedSize, numGen);
  failIfDoubleArrayNotEqual(&euc, expected, positions, expectedSize, 0.0);
}

static void runAllTests(void) {
  runTest(&euc, testImcaGenerateScanPositions01);
  runTest(&euc, testImcaGenerateScanPositions02);
  runTest(&euc, testImcaGenerateScanPositions03);
  runTest(&euc, testImcaGenerateScanPositions04);
  runTest(&euc, testImcaGenerateScanPositions05);
  runTest(&euc, testImcaGenerateScanPositions06);
  runTest(&euc, testImcaGenerateScanPositions07);
  runTest(&euc, testImcaGenerateScanPositions08);
  runTest(&euc, testImcaGenerateScanPositions09);
  runTest(&euc, testImcaGenerateScanPositions10);
}

void regScanPosGenTestRunner(void) {
  runTests(&euc, runAllTests);
}

static const iocshArg *const regScanPosGenTestRunnerArgs[] = {};
static const iocshFuncDef regScanPosGenTestRunnerDef = {
  "regScanPosGenTestRunner", 0, regScanPosGenTestRunnerArgs };
static void regScanPosGenTestRunnerCall(const iocshArgBuf *args) {
  regScanPosGenTestRunner();
}
static void regScanPosGenTestRegistrar(void) {
  static int isFirstTime = 1;
  if (!isFirstTime) return;
  isFirstTime = 0;
  iocshRegister(&regScanPosGenTestRunnerDef, regScanPosGenTestRunnerCall);
}
epicsExportRegistrar(regScanPosGenTestRegistrar);
